SystemOrganization addCategory: #'MC2-Repositories-Tests'!
SystemOrganization addCategory: #'MC2-Repositories-File'!

MDRepository subclass: #MDFileRepository
	instanceVariableNames: 'dataStream index memos lastPosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repositories-File'!

!MDFileRepository class methodsFor: 'accessing' stamp: 'cwp 10/12/2005 21:30'!
extension
	^ '.mcr'! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'cwp 2/2/2006 22:20'!
named: aString
	^ self on: (FileStream fileNamed: aString, self extension)! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'dc 7/19/2007 10:41'!
named: aString do: aBlock
	^ self
		on: (FileStream fileNamed: aString, self extension)
		do: aBlock! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'dc 6/14/2007 06:50'!
on: aFileStream
	^ self new
		initializeWithStream: aFileStream;
		yourself! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'dc 7/19/2007 10:41'!
on: aFileStream do: aBlock
	|repos|
	repos := self new
				initializeWithStream: aFileStream;
				yourself.
	^ [aBlock value: repos]
		ensure: [repos close]! !

!MDFileRepository methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:22'!
= other
	^ self species = other species
		and: [other dataStreamIs: dataStream]! !

!MDFileRepository methodsFor: 'synchronization' stamp: 'cwp 12/23/2007 00:11'!
allHashstamps
	self rescanData.
	^ index keys collect: [:ea | MDHashstamp fromBytes: ea]! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 10/13/2005 23:01'!
close
	dataStream close! !

!MDFileRepository methodsFor: 'private' stamp: 'cwp 12/25/2006 15:23'!
dataStreamIs: aStream
	^ dataStream = aStream! !

!MDFileRepository methodsFor: 'public' stamp: 'avi 3/20/2006 00:13'!
description
	^ dataStream stream localName! !

!MDFileRepository methodsFor: 'public' stamp: 'avi 4/19/2005 01:50'!
get: aHashstamp 
	| key bytes position |
	key := aHashstamp bytes.
	position := index at: key ifAbsent: [^self rescanData ifTrue: [self get: aHashstamp]].
	bytes := dataStream valueAtPosition: position.
	^MDMarshal materialize: bytes! !

!MDFileRepository methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:25'!
hash
	^ dataStream hash! !

!MDFileRepository methodsFor: 'synchronization' stamp: 'cwp 12/23/2007 00:12'!
includes: aHashstamp
	self rescanData.
	^ index includesKey: aHashstamp bytes! !

!MDFileRepository methodsFor: 'initialize-release' stamp: 'dc 6/14/2007 06:48'!
initializeWithStream: aFileStream
	dataStream := MDKeyedStream on: aFileStream.
	lastPosition := 0.
	index := Dictionary new.
	memos := Set new.
	self rescanData! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 1/19/2007 20:57'!
memos
	
	self rescanData.
	^ memos! !

!MDFileRepository methodsFor: 'public' stamp: 'dc 6/14/2007 07:18'!
put: aHashedObject
	| key value |
	key := aHashedObject hashstamp bytes.
	index at: key ifAbsent:
		[value := MDMarshal serialize: aHashedObject.
		index at: key put: dataStream nextValuePosition.
		dataStream appendKey: key value: value.
		dataStream flush]! !

!MDFileRepository methodsFor: 'private' stamp: 'cwp 1/19/2007 20:56'!
rescanData
	dataStream reopen.
	dataStream fromPosition: lastPosition keysValuesAndPositionsDo:  
		[:key :val :pos |
		index at: key put: pos.
		(MDMarshal classFor: val) = MDMemo ifTrue:
			[memos add: (MDMarshal materialize: val)]].
	dataStream position = lastPosition
		ifTrue: [^ false]
		ifFalse: [lastPosition _ dataStream position. ^ true]! !

TestCase subclass: #MDByteIndexTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repositories-Tests'!

!MDByteIndexTest methodsFor: 'as yet unclassified' stamp: 'avi 3/13/2005 04:01'!
storeAndRetrieveWithFileSize: aNumber
	|  file stream |
	stream _ RWBinaryOrTextStream on: ByteArray new.
	stream next: aNumber put: 0.
	stream reset.
 
	file _ MDByteIndex stream: stream keySize: 2 valueSize: 4.
	file at: 'aa' asByteArray put: 'abcd' asByteArray.
	file at: 'ab' asByteArray put: 'qxry' asByteArray.
	file at: 'xy' asByteArray put: '1234' asByteArray.
	self assert: file tally = 3.

	self assert: (file at: 'aa' asByteArray) = 'abcd' asByteArray.
	self assert: (file at: 'ab' asByteArray) = 'qxry' asByteArray.
	self assert: (file at: 'xy' asByteArray) = '1234' asByteArray.
	file at: 'ab' asByteArray put: '9876' asByteArray.
	self assert: (file at: 'ab' asByteArray) = '9876' asByteArray.
	self assert: file tally = 3! !

!MDByteIndexTest methodsFor: 'as yet unclassified' stamp: 'avi 3/13/2005 03:58'!
testStoreAndRetrieve
	self storeAndRetrieveWithFileSize: 100.
	self storeAndRetrieveWithFileSize: 24.
	self should: [self storeAndRetrieveWithFileSize: 12] raise: Error! !

TestCase subclass: #MDKeyedStreamTest
	instanceVariableNames: 'dataStream baseStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repositories-Tests'!

!MDKeyedStreamTest methodsFor: 'support' stamp: 'cwp 4/10/2005 16:24'!
append: aSmallInteger
	dataStream 
		appendKey: (self key: aSmallInteger)
		value: (self value: aSmallInteger).
! !

!MDKeyedStreamTest methodsFor: 'support' stamp: 'cwp 4/10/2005 16:19'!
assert: bytes is: aSmallInteger
	self assert: bytes last = aSmallInteger! !

!MDKeyedStreamTest methodsFor: 'support' stamp: 'cwp 4/10/2005 16:02'!
bytes: count containing: aSmallInteger
	| key |
	key _ ByteArray new: count.
	key at: key size put: aSmallInteger.
	^ key! !

!MDKeyedStreamTest methodsFor: 'support' stamp: 'cwp 4/10/2005 16:02'!
key: aSmallInteger
	^ self bytes: dataStream keySize containing: aSmallInteger! !

!MDKeyedStreamTest methodsFor: 'running' stamp: 'avi 4/17/2005 01:19'!
setUp
	baseStream := RWBinaryOrTextStream on: (ByteArray new: 100).
	dataStream := MDKeyedStream on: baseStream! !

!MDKeyedStreamTest methodsFor: 'tests' stamp: 'cwp 4/10/2005 16:11'!
testAppend
	self append: 1.
	dataStream keysValuesAndPositionsDo: 
		[:k :v :p |
		self assert: k is: 1.
		self assert: v is: 1.
		self assert: p >= 0]! !

!MDKeyedStreamTest methodsFor: 'tests' stamp: 'cwp 4/10/2005 16:18'!
testKeysValuePositions
	| index position |
	1 to: 3 do: [:i | self append: i].
	index _ 1.
	position _ -1.
	dataStream keysValuesAndPositionsDo: 
		[:k :v :p |
		self assert: k is: index.
		self assert: v is: index.
		self assert: p > position.
		position _ p.
		index _ index + 1]! !

!MDKeyedStreamTest methodsFor: 'tests' stamp: 'avi 4/17/2005 01:39'!
testPosition
	| position |
	position _ dataStream nextValuePosition.
	self append: 1.
	dataStream keysValuesAndPositionsDo: 
		[:k :v :p | self assert: p = position].
! !

!MDKeyedStreamTest methodsFor: 'tests' stamp: 'cwp 4/10/2005 17:17'!
testValueAtPosition
	| position |
	self append: 1.
	dataStream keysValuesAndPositionsDo: 
		[:k :v :p | position _ p].
	self assert: (dataStream valueAtPosition: position) is: 1! !

!MDKeyedStreamTest methodsFor: 'support' stamp: 'cwp 4/10/2005 16:07'!
value: aSmallInteger
	^ self bytes: dataStream keySize * 2 containing: aSmallInteger! !

Object subclass: #MDByteIndex
	instanceVariableNames: 'keySize valueSize streamSize stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repositories-File'!

!MDByteIndex class methodsFor: 'as yet unclassified' stamp: 'avi 3/15/2005 03:08'!
stream: aStream keySize: keyNumber valueSize: valueNumber
	^ self stream: aStream keySize: keyNumber valueSize: valueNumber streamSize: aStream size! !

!MDByteIndex class methodsFor: 'as yet unclassified' stamp: 'avi 3/19/2005 02:41'!
stream: aStream keySize: keyNumber valueSize: valueNumber streamSize: streamNumber
	^ self basicNew setStream: aStream keySize: keyNumber valueSize: valueNumber streamSize: streamNumber! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:36'!
at: keyArray
	^ self at: keyArray ifAbsent: [self error: 'Could not find key ', keyArray printString]! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:36'!
at: keyArray ifAbsent: errorBlock
	keyArray size = keySize ifFalse: [self error: 'Key must be exactly ', keySize asString, ' bytes'].
	^ self basicAt: keyArray ifAbsent: errorBlock! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:37'!
at: keyArray put: valueArray
	keyArray size = keySize ifFalse: [self error: 'Key must be exactly ', keySize asString, ' bytes'].
	valueArray size = valueSize ifFalse: [self error: 'Value must be exactly ', valueSize asString, ' bytes'].
	^ self basicAt: keyArray put: valueArray! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:34'!
basicAt: keyArray ifAbsent: errorBlock
	| i key |
	i _ self scanFor: keyArray.
	key _ self keyBytesAtIndex: i.
	^ key = keyArray
		ifTrue: [self valueBytesAtIndex: i]
		ifFalse: errorBlock! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:42'!
basicAt: keyArray put: valueArray
	| i key |
	i _ self scanFor: keyArray.
	key _ self keyBytesAtIndex: i.
	key = keyArray ifFalse:
		[self writeTally: self tally + 1.
		self writeKeyBytes: keyArray atIndex: i].
	self writeValueBytes: valueArray atIndex: i.
	^ valueArray! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:45'!
hash: keyArray
	^ keyArray hash! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/15/2005 03:38'!
keyBytesAtIndex: i
	self seekTo: i.
	^ stream atEnd
		ifTrue: [ByteArray new: keySize]
		ifFalse: [stream next: keySize]! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:05'!
keySize
	^ keySize! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/15/2005 03:07'!
maxIndex
	^ streamSize - 4 // self recordSize! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:43'!
recordSize
	^ keySize + valueSize! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:33'!
scanFor: keyArray
	| start |
	start _ (self hash: keyArray) \\ self maxIndex.
	start to: self maxIndex do: [:i | (self scanFor: keyArray at: i) ifTrue: [^ i]].
	0 to: start - 1 do: [:i | (self scanFor: keyArray at: i) ifTrue: [^ i]].
	
	self error: 'Index is full'! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:29'!
scanFor: keyArray at: i
	| key |
	key _ self keyBytesAtIndex: i.
	^ key = keyArray or: [key allSatisfy: [:b | b = 0]]! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:42'!
seekTo: i
	stream position: i * self recordSize + 4! !

!MDByteIndex methodsFor: 'initialize-release' stamp: 'avi 3/19/2005 02:41'!
setStream: aStream keySize: keyNumber valueSize: valueNumber streamSize: streamNumber
	keySize _ keyNumber.
	valueSize _ valueNumber.
	streamSize _ streamNumber.
	stream _ aStream.
	stream binary! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/19/2005 02:51'!
streamSize
	^ streamSize! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/15/2005 03:23'!
tally
	stream reset.
	^ stream atEnd
		ifTrue: [0]
		ifFalse: [stream nextInt32]! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:41'!
valueBytesAtIndex: i
	self seekTo: i.
	stream skip: keySize.
	^ stream next: valueSize! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:05'!
valueSize
	^ valueSize! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:40'!
writeKeyBytes: keyBytes atIndex: i
	self seekTo: i.
	stream nextPutAll: keyBytes! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:42'!
writeTally: aNumber
	stream reset.
	stream nextInt32Put: aNumber! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:40'!
writeValueBytes: valueBytes atIndex: i
	self seekTo: i.
	stream skip: keySize.
	stream nextPutAll: valueBytes! !

Object subclass: #MDKeyedStream
	instanceVariableNames: 'stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repositories-File'!

!MDKeyedStream class methodsFor: 'as yet unclassified' stamp: 'avi 4/17/2005 01:19'!
on: aStream
	^ self basicNew initializeWithStream: aStream! !

!MDKeyedStream methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:25'!
= other
	^ self species = other species
		and: [self stream name = other stream name]! !

!MDKeyedStream methodsFor: 'public' stamp: 'cwp 11/6/2007 22:52'!
appendKey: keyBytes value: valueBytes 
	self assert: [keyBytes size = self keySize].
	stream
		setToEnd;
		nextPutAll: keyBytes;
		nextInt32Put: valueBytes size;
		nextPutAll: valueBytes! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'cwp 10/13/2005 23:00'!
close
	stream name = 'a stream' ifFalse: [stream close]! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/17/2005 01:58'!
flush
	stream flush! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/19/2005 01:49'!
fromPosition: aNumber keysValuesAndPositionsDo: aBlock 
	| position key |
	stream position: aNumber.
	[stream atEnd] whileFalse: 
			[key := stream next: self keySize.
			position := stream position.
			aBlock
				value: key
				value: self readValue
				value: position]! !

!MDKeyedStream methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:25'!
hash
	^ self stream name hash! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'avi 4/17/2005 01:19'!
initializeWithStream: aPositionableStream
	stream _ aPositionableStream.
	stream binary! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'cwp 4/10/2005 15:14'!
keySize
	^ 24! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/17/2005 02:11'!
keysValuesAndPositionsDo: aBlock 
	self fromPosition: 0 keysValuesAndPositionsDo: aBlock
			! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/17/2005 01:38'!
nextValuePosition
	^ stream size + self keySize! !

!MDKeyedStream methodsFor: 'public' stamp: 'cwp 4/10/2005 17:29'!
position
	^ stream position! !

!MDKeyedStream methodsFor: 'private' stamp: 'avi 4/17/2005 01:29'!
readValue
	| length |
	length := stream nextInt32.
	^ stream next: length.
	! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'avi 3/22/2006 19:23'!
reopen
	stream name = 'a stream' ifFalse: [stream reopen; binary]! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'cwp 4/10/2005 14:53'!
stream
	^ stream! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/17/2005 01:28'!
valueAtPosition: anInteger
	stream position: anInteger.
	^ self readValue! !

MDRepositoryTest subclass: #MDFileRepositoryTest
	instanceVariableNames: 'second'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repositories-Tests'!

!MDFileRepositoryTest methodsFor: 'running' stamp: 'cwp 12/23/2007 00:06'!
destroyRepository: aRepository named: aString
	aRepository close.
	FileDirectory default deleteFileNamed: (aString, '.mcr')! !

!MDFileRepositoryTest methodsFor: 'running' stamp: 'cwp 1/1/2007 10:02'!
newRepository
	^ MDFileRepository named: self repositoryFileName! !

!MDFileRepositoryTest methodsFor: 'running' stamp: 'cwp 1/1/2007 10:02'!
repositoryFileName
	^ 'test-file-repository'! !

!MDFileRepositoryTest methodsFor: 'running' stamp: 'cwp 12/23/2007 00:10'!
secondFileName
	^ 'test-file-repository2'! !

!MDFileRepositoryTest methodsFor: 'running' stamp: 'cwp 12/23/2007 00:08'!
secondRepository
	^ second _ MDFileRepository named: self secondFileName! !

!MDFileRepositoryTest methodsFor: 'running' stamp: 'cwp 12/23/2007 00:09'!
tearDown
	self destroyRepository: repository named: self repositoryFileName.
	second ifNotNil:
		[self destroyRepository: second named: self secondFileName]! !

!MDFileRepositoryTest methodsFor: 'tests' stamp: 'cwp 1/1/2007 10:04'!
testConcurrentAppend
	| version repository2 get |
	[repository2 _ self newRepository.
	version _ self version.
	repository put: version.
	self assert: (repository get: version hashstamp) isEquivalentTo: version.
	get _ repository2 get: version hashstamp.
	self assert: get notNil.
	self assert: get isEquivalentTo: version]
		ensure: [repository2 ifNotNil: [repository2 close]]
! !

!MDFileRepositoryTest methodsFor: 'tests' stamp: 'avi 4/17/2005 01:41'!
testReopen
	| version |
	version _ self version.
	repository put: version.
	self assert: (repository get: version hashstamp) isEquivalentTo: version.
	repository _ self newRepository.
	self assert: (repository get: version hashstamp) isEquivalentTo: version.
! !
